/* Copyright 2015, 2016, 2017 Kevin Ryde

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   This file is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with this file.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Some C curve graph stuff using Nauty. */

#include <stdio.h>
#include <stdlib.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/naugroup.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"

/* GP-DEFINE  read("c-curve-defines.gp"); */

/* ------------------------------------------------------------------------- */
/* Misc Helpers */

/* Change (x,y) to (-y,x) which is a rotation by +90 degrees. */
#define C_CURVE_XY_ROTATE_PLUS90(x,y)           \
  do {                                          \
    int C_CURVE_XY_ROTATE_PLUS90__tmp = (x);    \
    (x) = -(y);                                 \
    (y) = C_CURVE_XY_ROTATE_PLUS90__tmp;        \
  } while (0)

/* Change (x,y) to (y,-x) which is a rotation by -90 degrees. */
#define C_CURVE_XY_ROTATE_MINUS90(x,y)          \
  do {                                          \
    int C_CURVE_XY_ROTATE_MINUS90__tmp = (x);   \
    (x) = (y);                                  \
    (y) = - C_CURVE_XY_ROTATE_MINUS90__tmp;     \
  } while (0)

/* dir(n) = CountOneBits(n) */
int c_curve_dir (int n)
{
  int dir = 0;
  while (n) {
    dir += (n&1);
    n >>= 1;
  }
  return dir;
}

int CountLowZeros (int n)
{
  int count = 0;
  while (n > 0 && (n & 1) == 0) { n >>= 1; count++; }
  return count;
}

/* turn(n) = 1 - CountLowZeros(n) */
int c_curve_turn (int n)
{
  return 1 - CountLowZeros(n);
}

static int dir4_to_dx[] = { 1,0,-1,0 };
static int dir4_to_dy[] = { 0,1,0,-1 };


/* ------------------------------------------------------------------------- */
/* C Curve Sequence Values */

void
c_curve_overflow (const char *str)
{
  fprintf(stderr, "C curve overflow: %s\n", str);
  abort();
}

/* c_curve_R(k) returns the right boundary length of C curve k.
   R(k) = [7,10]*2^floor(k/2)  - 2*k - 6
*/
int
c_curve_R (int k)
{
  int R = 1 << (k>>1);
  if (k & 1) {
    if ((R << 4) <= 0) { c_curve_overflow("R odd"); }
    R *= 10;
  } else {
    if ((R << 3) <= 0) { c_curve_overflow("R even"); }
    R *= 7;
  }
  return R - 2*k - 6;
}

/* Rpart(k) = 2^ceil(k/2) - vector_modulo([0,2],k) */
int
c_curve_Rpart (int k)
{
  return (1 << ((k+1)>>1)) - 2*(k&1);
}

int
c_curve_PartN (int k)
{
  int n = 1 << k;
  if (n <= 0) { c_curve_overflow("PartN"); }
  n += ((k&1) ? 1 : -1);
  if ((n % 3) != 0) {
    fprintf(stderr, "oops, k=%d not a multiple of 3: %d\n", k, n);
    abort();
  }
  return (n + ((n&1) ? 1 : -1))/3;
}

/* c_curve_V(k) returns the number of distinct points visited by C curve k.
   This is a 37 term recurrence so is included here just as values up to 32
   bits (signed).
*/
static int c_curve_V_array[] = {
  /* GP-VECTOR  c_curve_V_array */
  2,
  3,
  5,
  9,
  16,
  30,
  57,
  111,
  214,
  416,
  809,
  1579,
  3073,
  5995,
  11697,
  22853,
  44650,
  87342,
  170937,
  334839,
  656199,
  1286913,
  2525057,
  4957333,
  9736760,
  19133868,
  37616337,
  73986039,
  145578997,
  286570159,
  564326161,
  1111725765,
  /* GP-END */
};
/*
   GP-Test  poldegree(PolyV(x)) == 37
   GP-Test  vector(#c_curve_V_array,k,k--; V(k)) == c_curve_V_array
   GP-Test  #c_curve_V_array == 32
   for(k=0,100, my(n=V(k));if(n>2^31-1,break()); print(n,","));
*/
int
c_curve_V (int k)
{
  if (k >= sizeof(c_curve_V_array)/sizeof(c_curve_V_array[0])) {
    c_curve_overflow("V array");
  }
  return c_curve_V_array[k];
}

/* c_curve_Vpart(k) returns the number of distinct points visited by C curve part k.

         *   *         k=3  Vpart(3)=3
          \ /
           *

   *---*---*---*---*   k=4  Vpart(4)=6
           |
           *

   *   *   *   *   k=5  Vpart(5)=10
    \ / \ / \ /
     *   *   *
        / \
       *   *
        \ /
         *

*/
static int c_curve_Vpart_array[] = {
  /* GP-VECTOR  c_curve_Vpart_array */
  2,
  1,
  3,
  3,
  6,
  10,
  19,
  37,
  68,
  136,
  259,
  513,
  983,
  1941,
  3763,
  7395,
  14404,
  28290,
  55307,
  108597,
  212765,
  417951,
  820195,
  1612083,
  3167346,
  6229764,
  12252707,
  24116997,
  47475963,
  93515201,
  184240803,
  363158803,
  716008920,
  1412263118,
  /* GP-END */
};
/*
   GP-Test  vector(#c_curve_Vpart_array,k,k--; Vpart(k)) == c_curve_Vpart_array
   GP-Test  #c_curve_Vpart_array == 34
   for(k=0,100, my(n=Vpart(k));if(n>2^31-1,break()); print(n,","));
*/
int
c_curve_Vpart (int k)
{
  if (k >= sizeof(c_curve_Vpart_array)/sizeof(c_curve_Vpart_array[0])) {
    c_curve_overflow("Vpart array");
  }
  return c_curve_Vpart_array[k];
}

/* ------------------------------------------------------------------------- */
/* C Curve Graph */

/* c_curve_graph_n(k) returns the number of vertices in C curve k.
   This is the number of distinct points V(k) visited by the curve.

   c_curve_graph(g, m,n, k) sets graph g to C curve k.
   n should be the number of vertices as per c_curve_graph_n(k).
   g should have enough space for size m,n.
*/
int
c_curve_graph_n (int k, int part)
{
  return (part ? c_curve_Vpart(k) : c_curve_V(k));
}

/* global for experimenting */
DYNALLSTAT(int, v_to_x, v_to_x_sz);
DYNALLSTAT(int, v_to_y, v_to_y_sz);

void
c_curve_graph(graph *g, int m, int n, int k, int part)
{
  int i;
  int i_lo, i_hi;
  int from;
  int dx = 1;
  int dy = 0;
  int xy_upto = 1;
  int x = 0;
  int y = 0;

  if (part) {
    i_lo = c_curve_PartN(k);
    i_hi = c_curve_PartN(k+1);
    if (k > 0) {
      int dir4 = c_curve_dir(i_lo) & 3;
      dx = dir4_to_dx[dir4];
      dy = dir4_to_dy[dir4];
    }
  } else {
    i_lo = 0;
    i_hi = 1 << k;
    if (i_hi <= 0) { c_curve_overflow("2^k"); }
  }

  DYNALLOC1(int, v_to_x, v_to_x_sz, n, "malloc");
  DYNALLOC1(int, v_to_y, v_to_y_sz, n, "malloc");
  v_to_x[0] = x;
  v_to_y[0] = y;
  printf("start at xy=%d,%d dxdy=%d,%d\n", x,y, dx,dy);

  nauty_check(WORDSIZE, m,n, NAUTYVERSIONID);
  EMPTYGRAPH(g, m,n);

  from = 0;
  for (i = i_lo+1; i <= i_hi; i++) {
    int to, turn;
    x += dx;
    y += dy;
    for (to = 0; to < xy_upto; to++) {
      if (v_to_x[to] == x && v_to_y[to] == y) {
        goto GOT_TO;
      }
    }
    to = xy_upto++;
    if (to >= v_to_x_sz) {
      fprintf(stderr, "oops, c_curve_graph() i=%d(%d) to=%d exceeds size %d\n",
              i, i_hi, to, v_to_x_sz);
      abort();
    }
    v_to_x[to] = x;
    v_to_y[to] = y;
  GOT_TO:
    /* printf("  edge %d to %d at xy=%d,%d\n", from, to, x,y); */
    ADDONEEDGE(g, from, to, m);

    turn = c_curve_turn(i) & 3;
    switch (turn) {
    case 1: C_CURVE_XY_ROTATE_PLUS90(dx,dy);  break;
    case 2: dx=-dx; dy=-dy;                   break;
    case 3: C_CURVE_XY_ROTATE_MINUS90(dx,dy); break;
    default: break; /* turn==0 */
    }

    from = to;
  }

  if (xy_upto != n) {
    fprintf(stderr, "oops, c_curve_graph() k=%d xy_upto=%d but vertices n=%d\n",
            k, xy_upto, n);
    abort();
  }
  /* DYNFREE(v_to_x, v_to_x_sz); */
  /* DYNFREE(v_to_y, v_to_y_sz); */
}

/* ------------------------------------------------------------------------- */

void
writeautom (int *p, int n)
{
  int i;
  for (i = 0; i < n; i++)
    printf(" %2d",p[i]);
  printf("\n");
}

void
try_sequences (void)
{
  int k;
  for (k=0; k < 10; k++) {
    int R = c_curve_R(k);
    int V = c_curve_V(k);
    int start = c_curve_PartN(k);
    int end = c_curve_PartN(k+1);
    printf("k=%d  R=%d V=%d part %d to %d\n",
           k, R, V, start,end);
  }
  /*
    vector(10,k,k--; R(k))
  */
}

/* 1,4,20,120,631,3640,20204,118794,637899,3547428,19022140,105034148,560385768,3068832698,16391182512,89625624852,
 */
void
try_c_curve_Wiener (void)
{
  DYNALLSTAT(graph, g, g_sz);
  int k;
  int part = 1;

  for (k=0; k < 10; k++) {
    int n = c_curve_graph_n(k,part);
    int m = SETWORDSNEEDED(n);
    long long W_calc;

    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    c_curve_graph(g,m,n, k,part);
    struct nautyextra_Wiener_info info;
    nautyextra_Wiener_index(g, m,n, 0, &info);
    W_calc = info.W;

    /* printf("k=%d  Wiener index %Ld\n", k, W_calc); */
    printf("%Ld,", W_calc);
    fflush(stdout);
  }
}

/* C Curve diameter is right boundary */
void
try_c_curve_diameter (void)
{
  DYNALLSTAT(graph, g,g_sz);
  DYNALLSTAT(int, dist,dist_sz);
  int k;
  int part = 0;

  for (k=0; k < 10; k++) {
    int diameter_formula = c_curve_R(k);
    int n = c_curve_graph_n(k,part);
    int m = SETWORDSNEEDED(n);
    int radius = 0, diameter = 0, cycles = -1, u,v;
    long long W = 0;

    /* printf("k=%d  n V=%d\n", k, n); */

    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    c_curve_graph(g, m,n, k,part);
    diamstats(g, m,n, &radius, &diameter);
    if (n <= WORDSIZE) { cycles = cyclecount(g, m,n); }
    struct nautyextra_Wiener_info info;
    nautyextra_Wiener_index(g, m,n, 0, &info);
    W = info.W;

    DYNALLOC1(int, dist, dist_sz, n, "malloc");

    printf("k=%d vertices=%d  radius %d  diameter %d %d  cycles %d  W=%Ld\n",
           k, n, radius, diameter, diameter_formula, cycles, W);
    printf("  diameter attained by  ");
    for (u = 0; u < n; u++) {
      find_dist(g, m,n, u, dist);
      for (v = u; v < n; v++) {
        if (dist[v] == diameter) {
          printf(" %d-%d,", u,v);
        }
      }
    }
    printf("\n");
  }
}

/* part diameter 1,0,2,2,4,6,9,15,19,31,43,67,93,137,187,275,381
   radius        1,0,1,1,2,3,5, 8,12,20,28,41,58, 84,120,168,247
*/
void
try_c_curve_diameter_part (void)
{
  DYNALLSTAT(graph, g, g_sz);
  DYNALLSTAT(int, dist, dist_sz);
  int k;
  int part = 1;

  for (k=0; k < 10; k++) {
    int n = c_curve_graph_n(k,part);
    int m = SETWORDSNEEDED(n);
    int radius = 0, diameter = 0, cycles = -1, from,to;
    int Rpart = c_curve_Rpart(k);
    long long W = 0;
    unsigned long edges = 0;
    int mindeg=0, mincount=0, maxdeg=0, maxcount=0, eulerian=0;
    int segs = c_curve_PartN(k+1) - c_curve_PartN(k);

    /* printf("k=%d  n V=%d\n", k, n); */

    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    c_curve_graph(g, m,n, k,part);
    diamstats(g, m,n, &radius, &diameter);
    degstats(g, m,n, &edges, &mindeg, &mincount, &maxdeg, &maxcount, &eulerian);
    if (n <= WORDSIZE) { cycles = cyclecount(g, m,n); }
    /* W = nautyextra_Wiener_index(g, m,n); */

    printf("k=%d vertices=%d edges=%lu(%d)  radius %d  diameter %d Rpart %d  cycles %d  W=%Ld\n",
           k, n,edges,segs, radius, diameter, Rpart, cycles, W);

    DYNALLOC1(int, dist, dist_sz, n, "malloc");
    printf("  diameter attained by  ");
    for (from = 0; from < n; from++) {
      find_dist(g, m,n, from, dist);
      for (to = from; to < n; to++) {
        if (dist[to] == diameter) {
          printf(" %d-%d (%d,%d to %d,%d),",
                 from, to,
                 v_to_x[from],v_to_y[from],
                 v_to_x[to],v_to_y[to]);
        }
      }
    }
    printf("\n");
  }
}

/* ------------------------------------------------------------------------- */

int
main (void)
{
  setbuf(stdout,NULL);
  try_sequences();
  try_c_curve_Wiener();
  try_c_curve_diameter();
  try_c_curve_diameter_part();
  /* try_c_curve_graph_diameter(); */
  return 0;
}
